# -*- coding: utf-8 -*-

"""
Simulate shiny encounter rates

giovanni provides the following commands:

    Information
    ===========

    help      : display a list of available giovanni commands and their help
                information

    Simulations
    ===========

    simulate  : simulates a sequence of soft-resets for a given generation,
                producing a distribution of number of required resets


    Probability
    ===========

    odds      : calculate the probability of encountering a shiny by a given
                SR count

Additional help for each command can be accessed by using:

    giovanni <command> -h
"""

import sys

from giovanni.cli import odds, simulate, parsers, common
from giovanni import utils


def main():
    """Run the module"""
    p = parsers.generate_parser()
    args = p.parse_args()

    # Use debug mode if verbose is set
    level = common.get_logging_level(args, "INFO")
    logger = utils.get_logger(__name__, level=level)

    mod_name = args.cmd
    logger.debug(f"Calling giovanni.cli.{mod_name}.do_call")

    # Determine which command was run, delegate to its submodule
    if mod_name == "odds":
        exit_code = odds.do_call(args)
    elif mod_name == "simulate":
        exit_code = simulate.do_call(args)
    else:
        raise RuntimeError(
            f"Internal error, unknown command: {mod_name}. Please report this "
            f"as an issue to the developers!"
        )

    sys.exit(exit_code)



# -*- coding: utf-8 -*-

"""
Get odds for the generations

TODO: Wishlist
  . Support roman numerals for generations
  . Masuda method
  . Radar rates
"""

__all__ = [
    'get_base_odds',
]

_base_rates = {
    1: 0.,
    2: 1. / 8192.,
    3: 1. / 8192.,
    4: 1. / 8192.,
    5: 1. / 8192.,
    6: 1. / 4096.,
    7: 1. / 4096.,
}

_charm_multiplier = {
    1: 1.,
    2: 1.,
    3: 1.,
    4: 1.,

    # not present in 1-4
    5: 2.,
    6: 2.,
    7: 2.,
}


def get_base_odds(gen, charm):
    """Get the shiny odds for a generation

    Parameters
    ----------
    gen : int
        The generation/version being played. There are different odds for each.

    charm : bool
        Whether the shiny charm is equipped

    Returns
    -------
    odds : float
        The probability that any SR will result in a shiny appearing.
    """
    if gen not in _base_rates:
        raise ValueError(f"Unknown generation: {gen}")

    multiplier = 1.
    if charm:
        multiplier = _charm_multiplier[gen]

    return _base_rates[gen] * multiplier



# -*- coding: utf-8 -*-

"""
Main module and argparser for the `simulate` command

    $ giovanni simulate <options>
"""

import numpy as np
import pandas as pd

import textwrap
import tqdm

from giovanni.cli import common
from giovanni import odds
from giovanni import utils


def configure_subparser(subparsers):
    """Create the parser for the `simulate` commands"""
    description = textwrap.dedent("""
    Simulates the number of soft resets needed to encounter a shiny, producing
    a plot and statistics.
    """)

    example = textwrap.dedent("""
    Examples:

        $ giovanni simulate --gen 4 -n 100
    """)

    p = subparsers.add_parser(
        "simulate",
        description=description,
        help=description,
        epilog=example,
    )

    _, optional_args = common.get_argument_groups(p, "simulate")
    optional_args.add_argument(
        "-n", "--num",
        dest="num",
        type=int,
        help="The number of simulations to perform. If not provided, will "
             "default to 1000",
    )

    p.set_defaults(
        num=1000,
    )


def do_call(args):
    """Call the `simulate` CLI entrypoint"""
    level = common.get_logging_level(args, "INFO")
    logger = utils.get_logger(__name__, level=level)

    gen = common.parse_generation(args)

    num_sim = args.num
    swarm_size = args.swarm_size
    charm = args.charm

    if num_sim < 1:
        raise ValueError("--num must be a positive integer")

    logger.debug(f"Num simulations: {num_sim}")
    logger.debug(f"Swarm size: {swarm_size}")
    logger.debug(f"Shiny charm: {charm}")

    base_rate = odds.get_base_odds(gen, charm)
    logger.debug(f"Generation: {gen} (base rate={base_rate:.6f})")

    simulations = []
    for _ in tqdm.tqdm(range(num_sim), desc="simulation"):
        simulations.append(_simulation(base_rate, swarm_size))

    simulations = np.asarray(simulations)
    name = f"Required encounters (gen={gen}, " \
           f"num_trials={num_sim}, " \
           f"swarm_size={swarm_size})"

    table = pd.Series({
        "Average": np.round(np.average(simulations), 3),
        "Std Dev": np.round(np.std(simulations), 3),
        "Max": np.max(simulations),
        "Min": np.min(simulations),
    }, name=name)

    logger.info(
        f"Montecarlo simulation results --\n{table}"
    )

    return 0


def _simulation(rate, swarm_size, batch_size=10000):
    """Calculate the number of SRs required for a single simulation"""
    arange = np.arange(batch_size)
    rs = np.random.RandomState()

    # TODO: cython would be cool to speed this up
    def _sim_batch(n):
        """
        Generate a batch of random numbers. It is faster to compute
        randoms in a batch than it is one by one. Once this random generator
        is exhausted, generate a new one.
        """
        batch = rs.rand(n, swarm_size)
        mask = np.any(batch < rate, axis=1)

        if not mask.any():
            return None

        # Get the first scenario it finds
        return arange[mask][0]

    i = 1
    while True:
        which_idx = _sim_batch(batch_size)

        if which_idx is None:
            i += batch_size
        else:
            i += which_idx
            return i



# -*- coding: utf-8 -*-

"""
Configure the argument parsers
"""

import argparse

import giovanni
from giovanni.cli import odds, simulate


def generate_parser():
    """Get the CLI parser

    Configures a parser and attaches subparsers for all CLI commands. Note that
    every new command will require a new subparser. Subparsers are located in
    their own submodule, i.e.::

        odds.py
        simulate.py

    Each submodule should define a `configure_subparser` method which is called
    to add subparser metadata and arguments. Each subparser's namespace tells
    the `do_call` arbitration logic which command is being run.
    """
    p = argparse.ArgumentParser(
        description="giovanni is a tool for simulating shiny encouter rates "
                    "or computing the odds of a shiny encounter"
    )
    p.add_argument(
        "-V", "--version",
        action="version",
        version=f"giovanni {giovanni.__version__}",
        help="Show the giovanni version number and exit.",
    )

    subparsers = p.add_subparsers(
        metavar="commands",
        dest="cmd",
    )
    # http://bugs.python.org/issue9253
    # http://stackoverflow.com/a/18283730/1599393
    subparsers.required = True

    odds.configure_subparser(subparsers)
    simulate.configure_subparser(subparsers)

    return p



# -*- coding: utf-8 -*-

"""
Main module and argparser for the probability/odds commands

    $ giovanni odds <options>
"""

import textwrap

from giovanni.cli import common
from giovanni import odds
from giovanni import utils


def configure_subparser(subparsers):
    """Create the parser for the `odds` commands"""
    description = textwrap.dedent("""
    Calculate the odds of having encountered a shiny after a given number of
    soft-resets.
    """)

    example = textwrap.dedent("""
    Examples:

        $ giovanni odds --gen 4 --soft_resets 100
    """)

    p = subparsers.add_parser(
        "odds",
        description=description,
        help=description,
        epilog=example,
    )

    _, optional_args = common.get_argument_groups(p, "odds")
    optional_args.add_argument(
        "-r", "--soft_resets",
        dest="soft_resets",
        type=int,
        help="The number of soft resets to use in the odds calculation. If "
             "not provided, will default to 1",
    )

    p.set_defaults(
        soft_resets=1,
    )


def do_call(args):
    """Call the `odds` CLI entrypoint"""
    level = common.get_logging_level(args, "INFO")
    logger = utils.get_logger(__name__, level=level)

    gen = common.parse_generation(args)

    num_sr = args.soft_resets
    swarm_size = args.swarm_size
    charm = args.charm

    logger.debug(f"Num soft resets: {num_sr}")
    logger.debug(f"Swarm size: {swarm_size}")
    logger.debug(f"Shiny charm: {charm}")

    base_rate = odds.get_base_odds(gen, charm)
    logger.debug(f"Generation: {gen} (base rate={base_rate:.6f})")

    proba = 100 * _get_proba(
        base_rate, num_sr, swarm_size=swarm_size, charm=charm,
    )

    logger.info(
        f"Probability of shiny encounter after {num_sr:,} soft "
        f"reset{'s' if num_sr > 1 else ''}: {proba:.3f}%"
    )

    return 0


def _get_proba(rate, n_tries, swarm_size, charm):
    """Compute the probability"""
    p_no_encounter_single = (1. - rate)
    p_no_encouter = p_no_encounter_single ** n_tries
    return 1 - p_no_encouter



# -*- coding: utf-8 -*-

import logging
import logging.config

import copy


_default_logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "[%(levelname)s] %(asctime)s [%(name)s] - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S"
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        },
    },
    "root": {
        "level": "INFO",
        "handlers": ["console"]
    }
}


def get_logger(name, level="INFO", config=None):
    """Return a configured logger

    References
    ----------
    .. [1] https://docs.python.org/3/library/logging.html
    """
    if config is None:
        config = copy.deepcopy(_default_logging_config)

        # set the level in the default
        level = level.upper()
        config["handlers"]["console"]["level"] = level
        config["root"]["level"] = level

    logging.config.dictConfig(config)
    return logging.getLogger(name)



# -*- coding: utf-8 -*-

import numpy as np
from unittest import mock
from giovanni.cli import simulate


class MockRandomState:
    v = None

    def rand(self, *_):
        return self.v


def test_sim_batch_first():
    rate = 1. / 8192.
    swarm_size = 1
    batch_size = 10000

    with mock.patch("numpy.random.RandomState", MockRandomState) as mrs:
        mrs.v = 1e-12 * np.ones((batch_size, swarm_size))
        actual_encounter = simulate._simulation(rate, swarm_size, batch_size)

    assert 1 == actual_encounter


def test_sim_batch_last():
    rate = 1. / 8192.
    swarm_size = 1
    batch_size = 10000

    with mock.patch("numpy.random.RandomState", MockRandomState) as mrs:
        mrs.v = np.ones((batch_size, swarm_size))
        mrs.v[-1, 0] = 1e-12

        actual_encounter = simulate._simulation(rate, swarm_size, batch_size)

    assert batch_size == actual_encounter
